#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Remote helper utilities."""
import logging

from .lxc import LXC

logger = logging.getLogger(__name__)


BUILDD_REMOTE_NAME = "craft-com.ubuntu.cloud-buildd"
BUILDD_REMOTE_ADDR = "https://cloud-images.ubuntu.com/buildd/releases"


def configure_buildd_image_remote(
    lxc: LXC = LXC(),
) -> str:
    """Configure buildd remote, adding remote as required.

    :param lxc: LXC client.

    :returns: Name of remote to pass to launcher.
    """
    if BUILDD_REMOTE_NAME in lxc.remote_list():
        logger.debug("Remote %r already exists.", BUILDD_REMOTE_NAME)
    else:
        lxc.remote_add(
            remote=BUILDD_REMOTE_NAME,
            addr=BUILDD_REMOTE_ADDR,
            protocol="simplestreams",
        )
        logger.debug("Remote %r was successfully added.", BUILDD_REMOTE_NAME)

    return BUILDD_REMOTE_NAME



#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""API provider for Multipass.

This implementation interfaces with multipass using the `multipass` command-line
utility.
"""

import io
import json
import locale
import logging
import pathlib
import shlex
import subprocess
import time
from typing import Any, Callable, Dict, List, Optional, Tuple

import pkg_resources

from craft_providers import errors

from .errors import MultipassError

logger = logging.getLogger(__name__)


class Multipass:
    """Wrapper for multipass command.

    :param multipass_path: Path to multipass command to use.
    :cvar minimum_required_version: Minimum required version for compatibility.
    """

    minimum_required_version = "1.7"

    def __init__(
        self, *, multipass_path: pathlib.Path = pathlib.Path("multipass")
    ) -> None:
        self.multipass_path = multipass_path

    def _run(self, command: List[str], **kwargs) -> subprocess.CompletedProcess:
        """Execute a multipass command.

        It always checks the result (as no errors should pass silently) and captures the
        output (so `multipass` does not pollute the terminal).
        """
        command = [str(self.multipass_path), *command]

        logger.debug("Executing on host: %s", shlex.join(command))
        return subprocess.run(command, check=True, capture_output=True, **kwargs)

    def delete(self, *, instance_name: str, purge=True) -> None:
        """Passthrough for running multipass delete.

        :param instance_name: The name of the instance_name to delete.
        :param purge: Flag to purge the instance_name's image after deleting.

        :raises MultipassError: on error.

        """
        command = ["delete", instance_name]
        if purge:
            command.append("--purge")

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to delete VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def exec(
        self,
        *,
        command: List[str],
        instance_name: str,
        runner: Callable = subprocess.run,
        **kwargs,
    ):
        """Execute command in instance_name with specified runner.

        :param command: Command to execute in the instance.
        :param instance_name: Name of instance to execute in.
        :param runner: Execution function to invoke, e.g. subprocess.run or
            Popen.  First argument is finalized command with the attached
            kwargs.
        :param kwargs: Additional kwargs for runner.

        :returns: Runner's instance.
        """
        final_cmd = [str(self.multipass_path), "exec", instance_name, "--", *command]

        quoted_final_cmd = shlex.join(final_cmd)
        logger.debug("Executing on host: %s", quoted_final_cmd)

        return runner(final_cmd, **kwargs)  # pylint: disable=subprocess-run-check

    def info(self, *, instance_name: str) -> Dict[str, Any]:
        """Get information/state for instance.

        :returns: Parsed json data from info command.

        :raises MultipassError: On error.
        """
        command = ["info", instance_name, "--format", "json"]

        try:
            proc = self._run(command, text=True)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to query info for VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        return json.loads(proc.stdout)

    def is_supported_version(self) -> bool:
        """Check if Multipass version is supported.

        A helper to check if Multipass meets minimum supported version for
        craft-providers.

        :returns: True if installed version is supported.
        """
        version, _ = self.version()

        return pkg_resources.parse_version(version) >= pkg_resources.parse_version(
            self.minimum_required_version
        )

    def launch(
        self,
        *,
        instance_name: str,
        image: str,
        cpus: Optional[str] = None,
        mem: Optional[str] = None,
        disk: Optional[str] = None,
    ) -> None:
        """Launch multipass VM.

        :param instance_name: The name the launched instance will have.
        :param image: Name of image to create the instance with.
        :param cpus: Amount of virtual CPUs to assign to the launched instance.
        :param mem: Amount of RAM to assign to the launched instance.
        :param disk: Amount of disk space the launched instance will have.

        :raises MultipassError: on error.
        """
        command = ["launch", image, "--name", instance_name]
        if cpus is not None:
            command.extend(["--cpus", cpus])
        if mem is not None:
            command.extend(["--mem", mem])
        if disk is not None:
            command.extend(["--disk", disk])

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to launch VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def list(self) -> List[str]:
        """List names of VMs.

        :returns: Data from stdout if instance exists, else None.

        :raises MultipassError: On error.
        """
        command = ["list", "--format", "json"]

        try:
            proc = self._run(command, text=True)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief="Failed to query list of VMs.",
                details=errors.details_from_called_process_error(error),
            ) from error

        data_list = json.loads(proc.stdout).get("list", [])
        return [instance["name"] for instance in data_list]

    def mount(
        self,
        *,
        source: pathlib.Path,
        target: str,
        uid_map: Optional[Dict[str, str]] = None,
        gid_map: Optional[Dict[str, str]] = None,
    ) -> None:
        """Mount host source path to target.

        :param source: Path of local directory to mount.
        :param target: Target mount points, in <name>[:<path>] format, where
            <name> is an instance name, and optional <path> is the mount point.
            If omitted, the mount point will be the same as the source's
            absolute path.
        :param uid_map: A mapping of user IDs for use in the mount of the form
            <host-id> -> <instance-id>.  File and folder ownership will be
            mapped from <host-id> to <instance-id> inside the instance.
        :param gid_map: A mapping of group IDs for use in the mount of the form
            <host-id> -> <instance-id>.  File and folder ownership will be
            mapped from <host-id> to <instance-id> inside the instance.
        """
        command = ["mount", str(source), target]

        if uid_map is not None:
            for host_id, instance_id in uid_map.items():
                command.extend(["--uid-map", f"{host_id}:{instance_id}"])

        if gid_map is not None:
            for host_id, instance_id in gid_map.items():
                command.extend(["--gid-map", f"{host_id}:{instance_id}"])

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to mount {str(source)!r} to {target!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def start(self, *, instance_name: str) -> None:
        """Start VM instance.

        :param instance_name: the name of the instance to start.

        :raises MultipassError: on error.
        """
        command = ["start", instance_name]

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to start VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def stop(self, *, instance_name: str, delay_mins: int = 0) -> None:
        """Stop VM instance.

        :param instance_name: the name of the instance_name to stop.
        :param delay_mins: Delay shutdown for specified number of minutes.

        :raises MultipassError: on error.
        """
        command = ["stop", instance_name]

        if delay_mins != 0:
            command.extend(["--time", str(delay_mins)])

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to stop VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def transfer(self, *, source: str, destination: str) -> None:
        """Transfer to destination path with source IO.

        :param source: The source path, prefixed with <name:> for a path inside
            the instance.
        :param destination: The destination path, prefixed with <name:> for a
            path inside the instance.

        :raises MultipassError: On error.
        """
        command = ["transfer", source, destination]

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to transfer {source!r} to {destination!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def transfer_destination_io(
        self, *, source: str, destination: io.BufferedIOBase, chunk_size: int = 4096
    ) -> None:
        """Transfer from source file to destination IO.

        Note that this can't use std{in,out}=open(...) due to LP #1849753.

        :param source: The source path, prefixed with <name:> for a path inside
            the instance.
        :param destination: An IO stream to write to.
        :param chunk_size: Number of bytes to transfer at a time.  Defaults to
            4096.

        :raises MultipassError: On error.
        """
        command = [str(self.multipass_path), "transfer", source, "-"]
        with subprocess.Popen(
            command,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as proc:

            # Should never happen, but mypy/pyright makes noise.
            assert proc.stdout is not None
            assert proc.stderr is not None

            while True:
                data = proc.stdout.read(chunk_size)
                if not data:
                    break

                destination.write(data)

            # Take one read of stderr in case there is anything useful
            # for debugging an error.
            stderr = proc.stderr.read()

        if proc.returncode != 0:
            raise MultipassError(
                brief=f"Failed to transfer file {source!r}.",
                details=errors.details_from_command_error(
                    cmd=command, stderr=stderr, returncode=proc.returncode
                ),
            )

    def transfer_source_io(
        self, *, source: io.BufferedIOBase, destination: str, chunk_size: int = 4096
    ) -> None:
        """Transfer to destination path with source IO.

        Note that this can't use std{in,out}=open(...) due to LP #1849753.

        :param source: An IO stream to read from.
        :param destination: The destination path, prefixed with <name:> for a
            path inside the instance.
        :param chunk_size: Number of bytes to transfer at a time.  Defaults to
            4096.

        :raises MultipassError: On error.
        """
        command = [str(self.multipass_path), "transfer", "-", destination]
        with subprocess.Popen(
            command, stdin=subprocess.PIPE, stderr=subprocess.PIPE
        ) as proc:
            # Should never happen, but mypy/pyright makes noise.
            assert proc.stdin is not None
            assert proc.stderr is not None

            while True:
                data = source.read(chunk_size)
                if not data:
                    break

                proc.stdin.write(data)

            # Close stdin before reading stderr, otherwise read() will hang
            # because process is waiting for more data.
            proc.stdin.close()

            # Take one read of stderr in case there is anything useful
            # for debugging an error.
            stderr = proc.stderr.read()

        if proc.returncode != 0:
            raise MultipassError(
                brief=f"Failed to transfer file to destination {destination!r}.",
                details=errors.details_from_command_error(
                    cmd=command, stderr=stderr, returncode=proc.returncode
                ),
            )

    def umount(self, *, mount: str) -> None:
        """Unmount target in VM.

        :param mount: Mount point in <name>[:<path>] format, where <name> are
            instance names, and optional <path> are mount points.  If omitted,
            all mounts will be removed from the named instance.

        :raises MultipassError: On error.
        """
        command = ["umount", mount]

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to unmount {mount!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def wait_until_ready(
        self, *, retry_wait: float = 0.25, timeout: Optional[float] = None
    ) -> Tuple[str, Optional[str]]:
        """Wait until Multipass is ready (upon install/startup).

        :param retry_wait: Time to sleep between retries.
        :param timeout: Timeout in seconds.

        :returns: Tuple of parsed versions (multipass, multipassd).  multipassd
            may be None if Multipass is not ready and the timeout limit is reached.
        """
        if timeout is not None:
            deadline: Optional[float] = time.time() + timeout
        else:
            deadline = None

        while True:
            multipass_version, multipassd_version = self.version()

            if multipassd_version is not None:
                return (multipass_version, multipassd_version)

            if deadline is not None and time.time() >= deadline:
                break

            time.sleep(retry_wait)

        raise MultipassError(
            brief="Timed out waiting for Multipass to become ready.",
        )

    def version(self) -> Tuple[str, Optional[str]]:
        """Get multipass and multipassd versions.

        :returns: Tuple of parsed versions (multipass, multipassd).  multipassd
                  may be None if Multipass is not yet ready.
        """
        try:
            proc = self._run(["version"])
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief="Failed to check version.",
                details=errors.details_from_called_process_error(error),
            ) from error

        try:
            output = proc.stdout.decode(encoding=locale.getpreferredencoding())
        except UnicodeDecodeError as error:
            raise MultipassError(
                brief="Failed to check version.",
                details=f"Failed to decode output: {proc.stdout!r}",
            ) from error

        # Expected multipass version output should look like:
        # * Scenario 1: multipassd not yet ready
        #
        #   multipass: 1.5.0
        #
        # * Scenario 2: typical Linux
        #
        #   multipass: 1.5.0
        #   multipassd: 1.5.0
        #
        # * Scenario 3: typical Mac
        #
        #   multipass: 1.5.0+mac
        #   multipassd: 1.5.0+mac
        #
        # * Scenario 4: typical Windows
        #
        #   multipass: 1.5.0+win
        #   multipassd: 1.5.0+win
        #
        # * Scenario 5: outdated Windows version with notice message
        #   See: https://github.com/canonical/multipass/issues/2020
        #
        #   multipass: 1.5.0+win
        #   multipassd: 1.5.0+win
        #
        #   SOME NOTICE INFORMATION....
        #
        # After stripping and splitting:
        #    - ['multipass', '1.5.0']
        #    - ['multipass', '1.5.0', 'multipassd', '1.5.0']
        #    - ['multipass', '1.5.0+mac', 'multipassd', '1.5.0+mac']
        #    - ['multipass', '1.5.0+win', 'multipassd', '1.5.0+win']
        #    - ['multipass', '1.5.0+win', 'multipassd', '1.5.0+win', ...]
        output_split = output.strip().split()
        if len(output_split) < 2 or output_split[0] != "multipass":
            raise MultipassError(
                brief=f"Unable to parse version output: {proc.stdout!r}",
            )

        multipass_version = output_split[1].split("+")[0]

        if len(output_split) >= 4 and output_split[2] == "multipassd":
            multipassd_version = output_split[3].split("+")[0]
        else:
            multipassd_version = None

        return (multipass_version, multipassd_version)



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Helper(s) for env command."""
import pathlib
from typing import Dict, List, Optional


def formulate_command(
    env: Optional[Dict[str, Optional[str]]] = None,
    *,
    chdir: Optional[pathlib.Path] = None,
    ignore_environment: bool = False,
) -> List[str]:
    """Create an env command with the specified environment.

    For each key-value, the env command will include the key=value argument to
    the env command.

    If a variable is None, then the env -u parameter will be used to unset it.

    An empty environment will simply yield the env command.

    NOTE: not all versions of `env` support --chdir, it is up to the caller to
    ensure compatibility.

    :param env: Environment flags to set/unset.
    :param chdir: Optional directory to run in.
    :param ignore_environment: Start with an empty environment.

    :returns: List of env command strings.
    """
    env_cmd = ["env"]

    if chdir:
        env_cmd.append(f"--chdir={chdir.as_posix()}")

    if ignore_environment:
        env_cmd.append("-i")

    if env is None:
        return env_cmd

    for key, value in env.items():
        if value is None:
            env_cmd += ["-u", key]
        else:
            env_cmd.append(f"{key}={value}")

    return env_cmd



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Persistent instance config / datastore resident in provided environment."""

import io
import pathlib
from typing import Any, Dict, Optional

import pydantic
import yaml

from craft_providers import Executor, errors
from craft_providers.util import temp_paths

from .errors import BaseConfigurationError


def update_nested_dictionaries(
    config_data: Dict[str, Any], new_data: Dict[str, Any]
) -> Dict[str, Any]:
    """Recursively update a dictionary containing nested dictionaries.

    New values are added and existing values are updated. No data are removed.

    :param config_data: dictionary of config data to update.
    :param new_data: data to update `config_data` with.
    """
    for key, value in new_data.items():
        if isinstance(value, dict):
            config_data[key] = update_nested_dictionaries(
                config_data.get(key, {}), value
            )
        else:
            config_data[key] = value
    return config_data


class InstanceConfiguration(pydantic.BaseModel, extra=pydantic.Extra.forbid):
    """Instance configuration datastore.

    :param compatibility_tag: Compatibility tag for instance.

    :param snaps: dictionary of snaps and their revisions, e.g.
      snaps:
        snapcraft:
          revision: "x100"
        charmcraft:
          revision: 834
    """

    compatibility_tag: Optional[str] = None
    snaps: Optional[Dict[str, Dict[str, Any]]] = None

    @classmethod
    def unmarshal(cls, data: Dict[str, Any]) -> "InstanceConfiguration":
        """Create and populate a new `InstanceConfig` object from dictionary data.

        The unmarshal method validates the data in the dictionary and populates
        the corresponding fields in the `InstanceConfig` object.

        :param data: The dictionary data to unmarshal.

        :return: The newly created `InstanceConfiguration` object.

        :raise BaseConfigurationError: If validation fails.
        """
        return InstanceConfiguration(**data)

    def marshal(self) -> Dict[str, Any]:
        """Create a dictionary containing the InstanceConfiguration data.

        :return: The newly created dictionary.
        """
        return self.dict(by_alias=True, exclude_unset=True)

    @classmethod
    def load(
        cls,
        executor: Executor,
        config_path: pathlib.Path = pathlib.Path("/etc/craft-instance.conf"),
    ) -> Optional["InstanceConfiguration"]:
        """Load an instance config file from an environment.

        :param executor: Executor for instance.
        :param config_path: Path to configuration file.
                            Default is `/etc/craft-instance.conf`.

        :return: The InstanceConfiguration object or None,
                 if the config does not exist or is empty.

        :raise BaseConfigurationError: If the file cannot be loaded from
                                       the environment.
        """
        with temp_paths.home_temporary_file() as temp_config_file:
            try:
                executor.pull_file(source=config_path, destination=temp_config_file)
            except errors.ProviderError as error:
                raise BaseConfigurationError(
                    brief=(
                        "Failed to read instance config"
                        f" in environment at {config_path}"
                    ),
                ) from error
            except FileNotFoundError:
                return None
            with open(temp_config_file, encoding="utf8") as file:
                data = yaml.safe_load(file)
                if data is None:
                    return None

                return cls.unmarshal(data)

    def save(
        self,
        executor: Executor,
        config_path: pathlib.Path = pathlib.Path("/etc/craft-instance.conf"),
    ) -> None:
        """Save an instance config file to an environment.

        :param executor: Executor for instance.
        :param config_path: Path to configuration file.
                            Default is `/etc/craft-instance.conf`.

        """
        data = self.marshal()

        executor.push_file_io(
            destination=config_path,
            content=io.BytesIO(yaml.dump(data).encode()),
            file_mode="0644",
        )

    @classmethod
    def update(
        cls,
        executor: Executor,
        data: Dict[str, Any],
        config_path: pathlib.Path = pathlib.Path("/etc/craft-instance.conf"),
    ) -> "InstanceConfiguration":
        """Update an instance config file in an environment.

        New values are added and existing values are updated. No data are removed.
        If there is no existing config to update, then a new config is created.

        :param executor: Executor for instance.
        :param data: The dictionary to update instance with.

        :return: The updated `InstanceConfiguration` object.
        """
        config_instance = cls.load(executor=executor, config_path=config_path)
        if config_instance is None:
            updated_config_instance = cls.unmarshal(data)
        else:
            updated_config_data = update_nested_dictionaries(
                config_data=config_instance.marshal(), new_data=data
            )
            updated_config_instance = InstanceConfiguration(**updated_config_data)

        updated_config_instance.save(executor=executor, config_path=config_path)

        return updated_config_instance



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Base errors."""

from typing import Optional

from craft_providers.errors import ProviderError


class BaseConfigurationError(ProviderError):
    """Error configuring the base."""


class BaseCompatibilityError(ProviderError):
    """Base configuration compatibility error.

    :param reason: Reason for incompatibility.
    """

    def __init__(self, reason: str, *, details: Optional[str] = None) -> None:
        self.reason = reason

        brief = f"Incompatible base detected: {reason}."
        resolution = "Clean incompatible instance and retry the requested operation."

        super().__init__(brief=brief, details=details, resolution=resolution)



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Multipass Instance."""
import io
import logging
import pathlib
import subprocess
from typing import Any, Dict, List, Optional

from craft_providers import errors
from craft_providers.util import env_cmd

from .. import Executor
from .errors import MultipassError
from .multipass import Multipass

logger = logging.getLogger(__name__)


def _rootify_multipass_command(
    command: List[str],
    *,
    cwd: Optional[pathlib.Path] = None,
    env: Optional[Dict[str, Optional[str]]] = None,
) -> List[str]:
    """Wrap a command to run as root with specified environment.

    - Use sudo to run as root (Multipass defaults to ubuntu user).
    - Configure sudo to set home directory.
    - Account for environment flags in env, if any.

    :param command: Command to execute.
    :param env: Additional environment flags to set.

    :returns: List of command strings for multipass exec.
    """
    sudo_cmd = ["sudo", "-H", "--"]

    if env is not None or cwd is not None:
        sudo_cmd += env_cmd.formulate_command(env, chdir=cwd)

    return sudo_cmd + command


class MultipassInstance(Executor):
    """Multipass Instance Lifecycle.

    :param name: Name of multipass instance.
    """

    def __init__(
        self,
        *,
        name: str,
        multipass: Optional[Multipass] = None,
    ):
        super().__init__()

        self.name = name

        if multipass is not None:
            self._multipass = multipass
        else:
            self._multipass = Multipass()

    def push_file_io(
        self,
        *,
        destination: pathlib.PurePath,
        content: io.BytesIO,
        file_mode: str,
        group: str = "root",
        user: str = "root",
    ) -> None:
        """Create or replace file with content and file mode.

        Multipass transfers data as "ubuntu" user, forcing us to first copy a
        file to a temporary location before moving to a (possibly) root-owned
        location and with appropriate permissions.

        :param destination: Path to file.
        :param content: Contents of file.
        :param file_mode: File mode string (e.g. '0644').
        :param group: File group owner/id.
        :param user: File user owner/id.
        """
        try:
            tmp_file_path = self._multipass.exec(
                instance_name=self.name,
                command=["mktemp"],
                runner=subprocess.run,
                capture_output=True,
                check=True,
                text=True,
            ).stdout.strip()

            self._multipass.transfer_source_io(
                source=content,
                destination=f"{self.name}:{tmp_file_path}",
            )

            self.execute_run(
                ["chown", f"{user}:{group}", tmp_file_path],
                capture_output=True,
                check=True,
            )

            self.execute_run(
                ["chmod", file_mode, tmp_file_path],
                capture_output=True,
                check=True,
            )

            self.execute_run(
                ["mv", tmp_file_path, destination.as_posix()],
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=(
                    f"Failed to create file {destination.as_posix()!r}"
                    f" in {self.name!r} VM."
                ),
                details=errors.details_from_called_process_error(error),
            ) from error

    def delete(self) -> None:
        """Delete instance and purge."""
        return self._multipass.delete(
            instance_name=self.name,
            purge=True,
        )

    def execute_popen(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.Path] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        **kwargs,
    ) -> subprocess.Popen:
        """Execute process in instance using subprocess.Popen().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param env: Additional environment to set for process.
        :param kwargs: Additional keyword arguments for subprocess.Popen().

        :returns: Popen instance.
        """
        return self._multipass.exec(
            instance_name=self.name,
            command=_rootify_multipass_command(command, cwd=cwd, env=env),
            runner=subprocess.Popen,
            **kwargs,
        )

    def execute_run(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.Path] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Execute command using subprocess.run().

        The process' environment will inherit the execution environment's
        default environment (PATH, etc.), but can be additionally configured via
        env parameter.

        :param command: Command to execute.
        :param env: Additional environment to set for process.
        :param kwargs: Keyword args to pass to subprocess.run().

        :returns: Completed process.

        :raises subprocess.CalledProcessError: if command fails and check is
            True.
        """
        return self._multipass.exec(
            instance_name=self.name,
            command=_rootify_multipass_command(command, cwd=cwd, env=env),
            runner=subprocess.run,
            **kwargs,
        )

    def exists(self) -> bool:
        """Check if instance exists.

        :returns: True if instance exists.

        :raises MultipassError: On unexpected failure.
        """
        vm_list = self._multipass.list()

        return self.name in vm_list

    def _get_info(self) -> Dict[str, Any]:
        """Get configuration and state for instance.

        :returns: State information parsed from multipass if instance exists,
            else None.

        :raises MultipassError: If unable to parse VM info.
        """
        info_data = self._multipass.info(instance_name=self.name).get("info")

        if info_data is None or self.name not in info_data:
            raise MultipassError(
                brief="Malformed multipass info",
                details=f"Returned data: {info_data!r}",
            )

        return info_data[self.name]

    def is_mounted(
        self, *, host_source: pathlib.Path, target: pathlib.PurePath
    ) -> bool:
        """Check if path is mounted at target.

        :param host_source: Host path to check.
        :param target: Instance path to check.

        :returns: True if host_source is mounted at target.

        :raises MultipassError: On unexpected failure.
        """
        info = self._get_info()
        mounts = info.get("mounts", {})

        for mount_point, mount_config in mounts.items():
            # Even on Windows, Multipass writes source_path as posix, e.g.:
            # 'C:/Users/chris/tmpbat91bwz.tmp-pytest'
            if (
                mount_point == target.as_posix()
                and mount_config.get("source_path") == host_source.as_posix()
            ):
                return True

        return False

    def is_running(self) -> bool:
        """Check if instance is running.

        :returns: True if instance is running.

        :raises MultipassError: On unexpected failure.
        """
        info = self._get_info()

        return info.get("state") == "Running"

    def launch(
        self,
        *,
        image: str,
        cpus: int = 2,
        disk_gb: int = 256,
        mem_gb: int = 2,
    ) -> None:
        """Launch instance.

        :param image: Name of image to create the instance with.
        :param instance_cpus: Number of CPUs.
        :param instance_disk_gb: Disk allocation in gigabytes.
        :param instance_mem_gb: Memory allocation in gigabytes.
        :param instance_name: Name of instance to use/create.
        :param instance_stop_time_mins: Stop time delay in minutes.

        :raises MultipassError: On unexpected failure.
        """
        self._multipass.launch(
            instance_name=self.name,
            image=image,
            cpus=str(cpus),
            disk=f"{disk_gb!s}G",
            mem=f"{mem_gb!s}G",
        )

    def mount(
        self,
        *,
        host_source: pathlib.Path,
        target: pathlib.PurePath,
    ) -> None:
        """Mount host host_source directory to target mount point.

        Checks first to see if already mounted.

        :param host_source: Host path to mount.
        :param target: Instance path to mount to.

        :raises MultipassError: On unexpected failure.
        """
        if self.is_mounted(host_source=host_source, target=target):
            return

        self._multipass.mount(
            source=host_source,
            target=f"{self.name}:{target.as_posix()}",
        )

    def pull_file(self, *, source: pathlib.PurePath, destination: pathlib.Path) -> None:
        """Copy a file from the environment to host.

        :param source: Environment file to copy.
        :param destination: Host file path to copy to.  Parent directory
            (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises MultipassError: On unexpected error copying file.
        """
        proc = self.execute_run(["test", "-f", source.as_posix()], check=False)
        if proc.returncode != 0:
            raise FileNotFoundError(f"File not found: {source.as_posix()!r}")

        if not destination.parent.is_dir():
            raise FileNotFoundError(f"Directory not found: {str(destination.parent)!r}")

        self._multipass.transfer(
            source=f"{self.name}:{source.as_posix()}", destination=str(destination)
        )

    def push_file(self, *, source: pathlib.Path, destination: pathlib.PurePath) -> None:
        """Copy a file from the host into the environment.

        The destination file is overwritten if it exists.

        :param source: Host file to copy.
        :param destination: Target environment file path to copy to.  Parent
            directory (destination.parent) must exist.

        :raises FileNotFoundError: If source file or destination's parent
            directory does not exist.
        :raises MultipassError: On unexpected error copying file.
        """
        if not source.is_file():
            raise FileNotFoundError(f"File not found: {str(source)!r}")

        proc = self.execute_run(
            ["test", "-d", destination.parent.as_posix()], check=False
        )
        if proc.returncode != 0:
            raise FileNotFoundError(
                f"Directory not found: {str(destination.parent.as_posix())!r}"
            )

        self._multipass.transfer(
            source=str(source),
            destination=f"{self.name}:{destination.as_posix()}",
        )

    def start(self) -> None:
        """Start instance.

        :raises MultipassError: On unexpected failure.
        """
        self._multipass.start(instance_name=self.name)

    def stop(self, *, delay_mins: int = 0) -> None:
        """Stop instance.

        :param delay_mins: Delay shutdown for specified minutes.

        :raises MultipassError: On unexpected failure.
        """
        self._multipass.stop(instance_name=self.name, delay_mins=delay_mins)

    def unmount(self, target: pathlib.Path) -> None:
        """Unmount mount target shared with host.

        :param target: Target shared with host to unmount.

        :raises MultipassError: On failure to unmount target.
        """
        mount = f"{self.name}:{target.as_posix()}"

        self._multipass.umount(mount=mount)

    def unmount_all(self) -> None:
        """Unmount all mounts shared with host.

        :raises MultipassError: On failure to unmount target.
        """
        self._multipass.umount(mount=self.name)



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Multipass Provider."""

import logging

from craft_providers import Base, bases

from .multipass_instance import MultipassInstance

logger = logging.getLogger(__name__)


def launch(
    name: str,
    *,
    base_configuration: Base,
    image_name: str,
    cpus: int = 2,
    disk_gb: int = 64,
    mem_gb: int = 2,
    auto_clean: bool = False,
) -> MultipassInstance:
    """Create, start, and configure instance.

    If auto_clean is enabled, automatically delete an existing instance that is
    deemed to be incompatible, rebuilding it with the specified environment.

    :param name: Name of instance.
    :param base_configuration: Base configuration to apply to instance.
    :param image_name: Multipass image to use, e.g. snapcraft:core20.
    :param cpus: Number of CPUs.
    :param disk_gb: Disk allocation in gigabytes.
    :param mem_gb: Memory allocation in gigabytes.
    :param auto_clean: Automatically clean instance, if incompatible.

    :returns: Multipass instance.

    :raises BaseConfigurationError: on unexpected error configuration base.
    :raises MultipassError: on unexpected Multipass error.
    """
    instance = MultipassInstance(name=name)

    if instance.exists():
        # TODO: Warn if existing instance doesn't match cpu/disk/mem specs.
        instance.start()
        try:
            base_configuration.warmup(executor=instance)
            return instance
        except bases.BaseCompatibilityError as error:
            if auto_clean:
                logger.debug(
                    "Cleaning incompatible instance %r (reason: %s).",
                    instance.name,
                    error.reason,
                )
                instance.delete()
            else:
                raise

    instance.launch(
        cpus=cpus,
        disk_gb=disk_gb,
        mem_gb=mem_gb,
        image=image_name,
    )
    base_configuration.setup(executor=instance)
    return instance



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Helpers for snap command."""

import pathlib
from typing import List


def formulate_ack_command(snap_assert_path: pathlib.Path) -> List[str]:
    """Formulate snap ack command to add assertions.

    :returns: List of ack command parts.
    """
    return ["snap", "ack", snap_assert_path.as_posix()]


def formulate_known_command(query: List[str]) -> List[str]:
    """Formulate snap known command to retrieve assertions.

    :returns: List of ack command parts.
    """
    return ["snap", "known", *query]


def formulate_local_install_command(
    classic: bool, dangerous: bool, snap_path: pathlib.Path
) -> List[str]:
    """Formulate snap install command.

    :param classic: Flag to enable installation of classic snap.
    :param dangerous: Flag to enable installation of snap without ack.

    :returns: List of command parts.
    """
    install_cmd = ["snap", "install", snap_path.as_posix()]

    if classic:
        install_cmd.append("--classic")

    if dangerous:
        install_cmd.append("--dangerous")

    return install_cmd


def formulate_pack_command(snap_name: str, output_file_path) -> List[str]:
    """Formulate the command to pack a snap from a directory.

    :param snap_name: The name of the channel.
    :param output_file_path: File path to the packed snap.

    :returns: List of command parts.
    """
    return [
        "snap",
        "pack",
        f"/snap/{snap_name}/current/",
        f"--filename={output_file_path}",
    ]


def formulate_remote_install_command(
    snap_name: str, channel: str, classic: bool
) -> List[str]:
    """Formulate the command to snap install from Store.

    :param snap_name: The name of the channel.
    :param channel: The channel to install the snap from.
    :param classic: Flag to enable installation of classic snap.
    :param dangerous: Flag to enable installation of snap without ack.

    :returns: List of command parts.
    """
    install_cmd = ["snap", "install", snap_name, "--channel", channel]

    if classic:
        install_cmd.append("--classic")

    return install_cmd


def formulate_refresh_command(snap_name: str, channel: str) -> List[str]:
    """Formulate snap refresh command.

    :param snap_name: The name of the channel.
    :param channel: The channel to install the snap from.

    :returns: List of command parts.
    """
    return ["snap", "refresh", snap_name, "--channel", channel]


def formulate_remove_command(snap_name: str) -> List[str]:
    """Formulate snap remove command.

    :param snap_name: The name of the channel.

    :returns: List of command parts.
    """
    return ["snap", "remove", snap_name]



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""LXD Errors."""

from typing import Optional

from craft_providers.errors import ProviderError


class LXDError(ProviderError):
    """Unexpected LXD error."""


class LXDInstallationError(LXDError):
    """LXD Installation Error.

    :param reason: Reason for install failure.
    :param details: Optional details to include.
    """

    def __init__(
        self,
        reason: str,
        *,
        details: Optional[str] = None,
    ) -> None:
        brief = f"Failed to install LXD: {reason}."
        resolution = (
            "Please visit https://linuxcontainers.org/lxd/getting-started-cli/"
            " for instructions on installing LXD for your operating system."
        )

        super().__init__(brief=brief, details=details, resolution=resolution)



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Helpers for snap commands."""

import contextlib
import json
import logging
import pathlib
import shlex
import subprocess
import urllib.parse
from typing import Dict, Iterator, List, Optional

import requests
import requests_unixsocket  # type: ignore

from craft_providers import Executor
from craft_providers.bases.instance_config import InstanceConfiguration
from craft_providers.errors import ProviderError, details_from_called_process_error
from craft_providers.util import snap_cmd, temp_paths

logger = logging.getLogger(__name__)


# possible sources for the snap (using these two constants instead
# of an enum because the values are persisted with JSON)
SNAP_SRC_HOST = "host"
SNAP_SRC_STORE = "store"


class SnapInstallationError(ProviderError):
    """Unexpected error during snap installation."""


def _download_host_snap(
    *, snap_name: str, output: pathlib.Path, chunk_size: int = 64 * 1024
) -> None:
    """Download the current host snap using snapd's APIs."""
    quoted_name = urllib.parse.quote(snap_name, safe="")
    url = f"http+unix://%2Frun%2Fsnapd.socket/v2/snaps/{quoted_name}/file"
    try:
        resp = requests_unixsocket.get(url)
    except requests.exceptions.ConnectionError as error:
        raise SnapInstallationError(
            brief="Unable to connect to snapd service."
        ) from error

    try:
        resp.raise_for_status()
    except requests.exceptions.HTTPError as error:
        raise SnapInstallationError(
            brief=f"Unable to download snap {snap_name!r} from snapd."
        ) from error

    with output.open("wb") as stream:
        for chunk in resp.iter_content(chunk_size):
            stream.write(chunk)


def _pack_host_snap(*, snap_name: str, output: pathlib.Path) -> None:
    """Pack the current host snap."""
    command = snap_cmd.formulate_pack_command(snap_name, output)

    logger.debug("Executing command on host: %s", shlex.join(command))
    subprocess.run(
        command,
        capture_output=True,
        check=True,
    )


def get_host_snap_info(snap_name: str) -> Dict[str, str]:
    """Get info about a snap installed on the host."""
    quoted_name = urllib.parse.quote(snap_name, safe="")
    url = f"http+unix://%2Frun%2Fsnapd.socket/v2/snaps/{quoted_name}"
    try:
        snap_info = requests_unixsocket.get(url)
    except requests.exceptions.ConnectionError as error:
        raise SnapInstallationError(
            brief="Unable to connect to snapd service."
        ) from error
    snap_info.raise_for_status()
    # TODO: represent snap info in a dataclass
    return snap_info.json()["result"]


def _get_target_snap_revision_from_snapd(
    snap_name: str, executor: Executor
) -> Optional[str]:
    """Get the revision of the snap on the target."""
    quoted_name = urllib.parse.quote(snap_name, safe="")
    url = f"http://localhost/v2/snaps/{quoted_name}"
    cmd = ["curl", "--silent", "--unix-socket", "/run/snapd.socket", url]
    try:
        proc = executor.execute_run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief="Unable to get target snap revision."
        ) from error

    result = json.loads(proc.stdout)
    if result["status-code"] == 404:
        # snap not found
        return None
    if result["status-code"] == 200:
        return result["result"]["revision"]
    raise SnapInstallationError(f"Unknown response from snapd: {result!r}")


def _get_snap_revision_ensuring_source(
    snap_name: str, source: str, executor: Executor
) -> Optional[str]:
    """Get revision of snap on target and ensure the installation source."""
    instance_config = InstanceConfiguration.load(executor=executor)
    if instance_config is None or instance_config.snaps is None:
        return None

    config = instance_config.snaps.get(snap_name)
    if config is None:
        # not installed before
        return None

    # use 'get' to retrieve the source to support configs
    # saved by previous versions of the lib
    if config.get("source") == source:
        # previously installed from specified source: ok
        return config["revision"]

    # installed from other source: remove it
    logger.debug(
        "Snap %r installed from other source (%s), removing", snap_name, config
    )
    cmd = snap_cmd.formulate_remove_command(snap_name)
    try:
        executor.execute_run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"Failed to inject snap {snap_name!r}.",
            details="unable to remove previously installed snap",
        ) from error
    return None


@contextlib.contextmanager
def _get_host_snap(snap_name: str) -> Iterator[pathlib.Path]:
    """Get snap installed on host containing the config.

    Snapd provides an API to fetch a snap. First use that to fetch a snap.
    If the snap is installed using `snap try`, it may fail to download. In
    that case, attempt to construct the snap by packing it ourselves.

    :yields: context manager that sets the temporary snap installation file
      as the target
    """
    with temp_paths.home_temporary_directory() as tmp_dir:
        snap_path = tmp_dir / f"{snap_name}.snap"
        try:
            _download_host_snap(snap_name=snap_name, output=snap_path)
        except SnapInstallationError:
            logger.warning(
                "Failed to fetch snap from snapd,"
                " falling back to `snap pack` to recreate"
            )
            _pack_host_snap(snap_name=snap_name, output=snap_path)

        yield snap_path


def _get_assertion(query: List[str]) -> bytes:
    """Get an assertion from snapd.

    :param query: assertion query to pass to `snap known`
    :returns: assertion data
    :raises SnapInstallationError: if 'snap known' call fails
    """
    command = snap_cmd.formulate_known_command(query=query)
    logger.debug("Executing command on host: %s", command)
    try:
        return subprocess.run(command, capture_output=True, check=True).stdout
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief="failed to get assertions for snap",
            details=details_from_called_process_error(error),
        ) from error


@contextlib.contextmanager
def _get_assertions_file(
    snap_name: str, snap_id: str, snap_revision: str
) -> Iterator[pathlib.Path]:
    """Get an assertion file for a snap.

    :param snap_name: Name of snap to inject
    :param snap_id: ID of the snap
    :param snap_revision: Revision of the snap

    :yields: context manager that will set the temporary snap assertion file
      as the target
    """
    logger.debug("Creating an assert file for snap %r", snap_name)
    assertion_queries = [
        [
            "account-key",
            "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v"
            "7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
        ],
        ["snap-declaration", f"snap-name={snap_name}"],
        ["snap-revision", f"snap-revision={snap_revision}", f"snap-id={snap_id}"],
    ]

    with temp_paths.home_temporary_file() as assert_file_path:
        with open(assert_file_path, "wb") as assert_file:
            for query in assertion_queries:
                assert_file.write(_get_assertion(query))
                assert_file.write(b"\n")
            assert_file.flush()
            yield assert_file_path


def _add_assertions_from_host(executor: Executor, snap_name: str) -> None:
    """Add assertions from the host into the target for a snap.

    :param executor: Executor for target
    :param snap_name: Name of snap to inject
    """
    target_assert_path = pathlib.Path(f"/tmp/{snap_name}.assert")
    snap_info = get_host_snap_info(snap_name)

    try:
        with _get_assertions_file(
            snap_name=snap_name,
            snap_id=snap_info["id"],
            snap_revision=snap_info["revision"],
        ) as host_assert_path:
            executor.push_file(
                source=host_assert_path,
                destination=target_assert_path,
            )
    except ProviderError as error:
        raise SnapInstallationError(
            brief=f"failed to copy assert file for snap {snap_name!r}",
            details="error copying snap assert file into target environment",
        ) from error

    try:
        executor.execute_run(
            snap_cmd.formulate_ack_command(snap_assert_path=target_assert_path),
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"failed to add assertions for snap {snap_name!r}",
            details=details_from_called_process_error(error),
        ) from error


def inject_from_host(*, executor: Executor, snap_name: str, classic: bool) -> None:
    """Inject snap from host snap.

    :param executor: Executor for target
    :param snap_name: Name of snap to inject
    :param classic: Install in classic mode
    :raises SnapInstallationError: on failure to inject snap
    """
    logger.debug("Installing snap %r from host (classic=%s)", snap_name, classic)

    host_revision = get_host_snap_info(snap_name)["revision"]
    target_revision = _get_snap_revision_ensuring_source(
        snap_name=snap_name,
        source=SNAP_SRC_HOST,
        executor=executor,
    )
    logger.debug("Revisions found: host=%r, target=%r", host_revision, target_revision)

    if target_revision is not None and target_revision == host_revision:
        logger.debug(
            "Skipping snap injection:"
            " target is already up-to-date with revision on host"
        )
        return

    target_snap_path = pathlib.Path(f"/tmp/{snap_name}.snap")
    is_dangerous = host_revision.startswith("x")

    if not is_dangerous:
        _add_assertions_from_host(executor=executor, snap_name=snap_name)

    with _get_host_snap(snap_name) as host_snap_path:
        try:
            executor.push_file(
                source=host_snap_path,
                destination=target_snap_path,
            )
        except ProviderError as error:
            raise SnapInstallationError(
                brief=f"failed to copy snap file for snap {snap_name!r}",
                details="error copying snap file into target environment",
            ) from error

    try:
        executor.execute_run(
            snap_cmd.formulate_local_install_command(
                classic=classic, dangerous=is_dangerous, snap_path=target_snap_path
            ),
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"failed to install snap {snap_name!r}",
            details=details_from_called_process_error(error),
        ) from error

    InstanceConfiguration.update(
        executor=executor,
        data={
            "snaps": {snap_name: {"revision": host_revision, "source": SNAP_SRC_HOST}}
        },
    )


def install_from_store(
    *, executor: Executor, snap_name: str, channel: str, classic: bool
) -> None:
    """Install snap from store into target.

    Perform installation using method which prevents refreshing.

    :param executor: Executor for target.
    :param snap_name: Name of snap to install.
    :param channel: Channel to install from.
    :param classic: Install in classic mode.

    :raises SnapInstallationError: on unexpected error.
    """
    logger.debug(
        "Installing snap %r from store (channel=%r, classic=%s)",
        snap_name,
        channel,
        classic,
    )
    target_revision = _get_snap_revision_ensuring_source(
        snap_name=snap_name,
        source=SNAP_SRC_STORE,
        executor=executor,
    )
    logger.debug("Revision found in target: %r", target_revision)

    if target_revision is None:
        # no snap present in the target environment, just install it
        cmd = snap_cmd.formulate_remote_install_command(
            snap_name=snap_name,
            channel=channel,
            classic=classic,
        )
    else:
        # refresh the already installed snap
        cmd = snap_cmd.formulate_refresh_command(
            snap_name=snap_name,
            channel=channel,
        )

    try:
        executor.execute_run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as error:
        raise SnapInstallationError(
            brief=f"Failed to install/refresh snap {snap_name!r}.",
            details=details_from_called_process_error(error),
        ) from error

    new_target_revision = _get_target_snap_revision_from_snapd(
        snap_name=snap_name,
        executor=executor,
    )
    logger.debug("Revision after install/refresh: %r", new_target_revision)

    InstanceConfiguration.update(
        executor=executor,
        data={
            "snaps": {
                snap_name: {"revision": new_target_revision, "source": SNAP_SRC_STORE}
            }
        },
    )



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""LXC wrapper."""
import enum
import logging
import pathlib
import shlex
import subprocess
from typing import Any, Callable, Dict, List, Optional

import yaml

from craft_providers import errors

from .errors import LXDError

logger = logging.getLogger(__name__)


# pylint: disable=too-many-lines


class StdinType(enum.Enum):
    """Mappings for input stream to pass to stdin for lxc commands."""

    INTERACTIVE = subprocess.DEVNULL
    NULL = None


def load_yaml(data):
    """Load yaml without additional resolvers.

    LXD may return YAML that has datetimes that are not valid when parsed to
    datetime.datetime().  Instead just use the base loader and avoid resolving
    this type (and others).
    """
    return yaml.load(data, Loader=yaml.BaseLoader)


class LXC:  # pylint: disable=too-many-public-methods
    """Wrapper for lxc command-line interface."""

    def __init__(
        self,
        *,
        lxc_path: pathlib.Path = pathlib.Path("lxc"),
    ):
        self.lxc_path = lxc_path

    def _run_lxc(
        self,
        command: List[str],
        *,
        check: bool,
        project: Optional[str] = None,
        stdin: StdinType = StdinType.INTERACTIVE,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Execute lxc command on host, allowing output to console.

        Handles the --project=project options if project is specified.
        :param command: lxc command to execute.
        :param check: Check if the lxc command exits with a non-zero exit code.
        :param project: Name of LXD project.
        :param stdin: What input stream to pass to lxc.
        :param kwargs: Additional parameters to pass to the lxc command.

        :returns: Completed process.
        """
        lxc_cmd = [str(self.lxc_path)]

        if project is not None:
            lxc_cmd += ["--project", project]

        lxc_cmd += command

        logger.debug("Executing on host: %s", shlex.join(lxc_cmd))

        # for subprocess, input takes priority over stdin
        if "input" in kwargs:
            return subprocess.run(lxc_cmd, check=check, **kwargs)

        return subprocess.run(lxc_cmd, check=check, stdin=stdin.value, **kwargs)

    def config_device_add_disk(
        self,
        *,
        instance_name: str,
        source: pathlib.Path,
        path: pathlib.PurePath,
        device: str,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Mount host source directory to target mount point.

        :param instance_name: Name of instance.
        :param source: Host path.
        :param path: Mount target in instance.
        :param device: Name of device.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = [
            "config",
            "device",
            "add",
            f"{remote}:{instance_name}",
            device,
            "disk",
            f"source={source.as_posix()}",
            f"path={path.as_posix()}",
        ]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to add disk to instance {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def config_device_remove(
        self,
        *,
        instance_name: str,
        device: str,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Mount host source directory to target mount point.

        :param instance_name: Name of instance.
        :param device: Name of device.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = [
            "config",
            "device",
            "remove",
            f"{remote}:{instance_name}",
            device,
        ]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to remove device from instance {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def config_device_show(
        self, *, instance_name: str, project: str = "default", remote: str = "local"
    ) -> Dict[str, Any]:
        """Show full device configuration.

        :param instance_name: Name of instance.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["config", "device", "show", f"{remote}:{instance_name}"]

        try:
            proc = self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to show devices for instance {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        return load_yaml(proc.stdout)

    def config_set(
        self,
        *,
        instance_name: str,
        key: str,
        value: str,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Set instance_name configuration key.

        :param instance_name: Name of instance.
        :param key: Config key name.
        :param value: Config key value.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["config", "set", f"{remote}:{instance_name}", key, value]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=(
                    f"Failed to set config key {key!r} to {value!r}"
                    f" for instance {instance_name!r}."
                ),
                details=errors.details_from_called_process_error(error),
            ) from error

    def delete(
        self,
        *,
        instance_name: str,
        force: bool = False,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Delete instance.

        :param instance_name: Name of instance.
        :param force: Force deletion if running.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["delete", f"{remote}:{instance_name}"]

        if force:
            command.append("--force")

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to delete instance {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def exec(
        self,
        *,
        command: List[str],
        instance_name: str,
        cwd: Optional[str] = None,
        mode: Optional[str] = None,
        project: str = "default",
        remote: str = "local",
        runner: Callable = subprocess.run,
        **kwargs,
    ):
        """Execute command in instance_name with specified runner.

        :param command: Command to execute in the instance.
        :param instance_name: Name of instance to execute in.
        :param cwd: Optional current working directory for command.
        :param mode: Override terminal mode Valid options include: "auto",
            "interactive", "non-interactive". lxd default is "auto".
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.
        :param runner: Execution function to invoke, e.g. subprocess.run or
            Popen.  First argument is finalized command with the attached
            kwargs.
        :param kwargs: Additional kwargs for runner.

        :returns: Runner's instance.
        """
        final_cmd = [
            str(self.lxc_path),
            "--project",
            project,
            "exec",
            f"{remote}:{instance_name}",
        ]

        if cwd is not None:
            final_cmd.extend(["--cwd", cwd])

        if mode is not None:
            final_cmd.extend(["--mode", mode])

        final_cmd += ["--", *command]

        logger.debug("Executing in container: %s", shlex.join(final_cmd))

        return runner(final_cmd, **kwargs)  # pylint: disable=subprocess-run-check

    def file_pull(
        self,
        *,
        instance_name: str,
        source: pathlib.PurePath,
        destination: pathlib.Path,
        create_dirs: bool = False,
        recursive: bool = False,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Retrieve file from instance_name.

        :param instance_name: Name of instance.
        :param source: Path in environment to pull.
        :param destination: Path in host to write to.
        :param create_dirs: Create any directories necessary.
        :param recursive: Recursively transfer files.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = [
            "file",
            "pull",
            f"{remote}:{instance_name}{source.as_posix()}",
            destination.as_posix(),
        ]

        if create_dirs:
            command.append("--create-dirs")

        if recursive:
            command.append("--recursive")

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=(
                    f"Failed to pull file {source.as_posix()!r}"
                    f" from instance {instance_name!r}."
                ),
                details=errors.details_from_called_process_error(error),
            ) from error

    def file_push(
        self,
        *,
        instance_name: str,
        source: pathlib.Path,
        destination: pathlib.PurePath,
        create_dirs: bool = False,
        recursive: bool = False,
        gid: Optional[int] = None,
        uid: Optional[int] = None,
        mode: Optional[str] = None,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Create file with content and file mode.

        :param instance_name: Name of instance to push file to.
        :param source: Path in host to push.
        :param destination: Path in environment to write to.
        :param create_dirs: Create any directories necessary.
        :param recursive: Recursively transfer files.
        :param gid: Optional gid to set on push (lxd's default is -1).
        :param uid: Optional uid to set on push (lxd's default is -1).
        :param mode: Optional file mode to set on file.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = [
            "file",
            "push",
            source.as_posix(),
            f"{remote}:{instance_name}{destination.as_posix()}",
        ]

        if create_dirs:
            command.append("--create-dirs")

        if recursive:
            command.append("--recursive")

        if mode is not None:
            command.append(f"--mode={mode}")

        if gid is not None:
            command.append(f"--gid={gid}")

        if uid is not None:
            command.append(f"--uid={uid}")

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=(
                    f"Failed to push file {source.as_posix()!r}"
                    f" to instance {instance_name!r}."
                ),
                details=errors.details_from_called_process_error(error),
            ) from error

    def has_image(
        self, image_name, *, project: str = "default", remote: str = "local"
    ) -> bool:
        """Check if image with given alias name is present.

        :param image_name: Name of image alias.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.
        """
        image_list = self.image_list(project=project, remote=remote)

        for image in image_list:
            for alias in image["aliases"]:
                if image_name == alias["name"]:
                    return True

        return False

    def info(
        self,
        *,
        instance_name: Optional[str] = None,
        project: str = "default",
        remote: str = "local",
    ) -> Dict[str, Any]:
        """Show instance or server information.

        :param instance_name: Optional instance name.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        if instance_name is None:
            instance_name = ""

        command = ["info", remote + ":" + instance_name]

        try:
            proc = self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to get info for remote {remote!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        try:
            return load_yaml(proc.stdout)
        except yaml.YAMLError as error:
            raise LXDError(
                brief="Failed to parse lxc info.",
                details=(
                    f"* Command that failed: {shlex.join(proc.args)!r}\n"
                    f"* Command output: {proc.stdout!r}"
                ),
            ) from error

    def launch(
        self,
        *,
        instance_name: str,
        image: str,
        image_remote: str,
        config_keys: Optional[Dict[str, str]] = None,
        ephemeral: bool = False,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Launch instance.

        :param instance_name: Name of instance to launch.
        :param image: Name of image to use.
        :param image_remote: Name of image's remote.
        :param config_keys: Configuration keys to set.
        :param ephemeral: Use ephemeral instance.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = [
            "launch",
            f"{image_remote}:{image}",
            f"{remote}:{instance_name}",
        ]

        if ephemeral:
            command.append("--ephemeral")

        if config_keys is not None:
            for config_key in [f"{k}={v}" for k, v in config_keys.items()]:
                command.extend(["--config", config_key])

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                stdin=StdinType.INTERACTIVE,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to launch instance {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def image_copy(
        self,
        *,
        image: str,
        image_remote: str,
        alias: Optional[str] = None,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Copy image.

        :param instance_name: Optional instance name.
        :param alias: New alias to add to image.
        :param image: Image to copy.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = [
            "image",
            "copy",
            f"{image_remote}:{image}",
            f"{remote}:",
        ]

        if alias is not None:
            command.append(f"--alias={alias}")

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to copy image {image!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def image_delete(
        self, *, image: str, project: str = "default", remote: str = "local"
    ) -> None:
        """Delete image.

        :param image: Image to delete.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = [
            "image",
            "delete",
            f"{remote}:{image}",
        ]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to delete image {image!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def image_list(
        self, *, project: str = "default", remote: str = "local"
    ) -> List[Dict[str, Any]]:
        """List images.

        :param project: Name of LXD project.
        :param remote: Name of LXD remote.
        """
        command = ["image", "list", f"{remote}:", "--format=yaml"]

        try:
            proc = self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to list images for project {project!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        try:
            return load_yaml(proc.stdout)
        except yaml.YAMLError as error:
            raise LXDError(
                brief="Failed to parse lxc image list.",
                details=(
                    f"* Command that failed: {shlex.join(proc.args)!r}\n"
                    f"* Command output: {proc.stdout!r}"
                ),
            ) from error

    def list(
        self,
        *,
        project: str = "default",
        remote: str = "local",
    ) -> List[Dict[str, Any]]:
        """List instances and their status.

        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :returns: List of containers and their info.

        :raises LXDError: on unexpected error.
        """
        command = ["list", f"{remote}:", "--format=yaml"]

        try:
            proc = self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to list instances for project {project!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        try:
            return load_yaml(proc.stdout)
        except yaml.YAMLError as error:
            raise LXDError(
                brief="Failed to parse lxc list.",
                details=(
                    f"* Command that failed: {shlex.join(proc.args)!r}\n"
                    f"* Command output: {proc.stdout!r}"
                ),
            ) from error

    def list_names(
        self, *, project: str = "default", remote: str = "local"
    ) -> List[str]:
        """List container names.

        A helper to get a list of container names from list().

        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :returns: List of container names.

        :raises LXDError: on unexpected error.
        """
        instances = self.list(project=project, remote=remote)

        try:
            return [i["name"] for i in instances]
        except KeyError as error:
            raise LXDError(
                brief="Failed to parse lxc list.",
                details=(f"* Data received from lxc list: {instances!r}"),
            ) from error

    def profile_edit(
        self,
        *,
        profile: str,
        config: Dict[str, Any],
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Set profile configuration.

        :param profile: Name of profile.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["profile", "edit", f"{remote}:{profile}"]
        encoded_config = yaml.dump(config).encode()

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
                input=encoded_config,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to set profile {profile!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def profile_show(
        self, *, profile: str, project: str = "default", remote: str = "local"
    ) -> Dict[str, Any]:
        """Get profile configuration.

        :param profile: Name of profile.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["profile", "show", f"{remote}:{profile}"]

        try:
            proc = self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to show profile {profile!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        return load_yaml(proc.stdout)

    def project_create(self, *, project: str, remote: str = "local") -> None:
        """Create project.

        :param project: Name of LXD project to create.
        :param remote: Name of LXD remote to create project on.

        :raises LXDError: on unexpected error.
        """
        command = ["project", "create", f"{remote}:{project}"]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to create project {project!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def project_delete(self, *, project: str, remote: str = "local") -> None:
        """Delete project, if it exists.

        :param project: Name of LXD project to delete.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["project", "delete", f"{remote}:{project}"]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to delete project {project!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def project_list(self, remote: str = "local") -> List[str]:
        """Get list of projects.

        :param remote: Name of LXD remote to query.

        :returns: List of project names.

        :raises LXDError: on unexpected error.
        """
        command = ["project", "list", f"{remote}:", "--format=yaml"]

        try:
            proc = self._run_lxc(
                command,
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to list projects on remote {remote!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        try:
            projects = load_yaml(proc.stdout)
            return sorted([p["name"] for p in projects])
        except (KeyError, yaml.YAMLError) as error:
            raise LXDError(
                brief="Failed to parse lxc project list.",
                details=(
                    f"* Command that failed: {shlex.join(proc.args)!r}\n"
                    f"* Command output: {proc.stdout!r}"
                ),
            ) from error

    def publish(
        self,
        *,
        instance_name: str,
        alias: Optional[str] = None,
        force: bool = False,
        image_remote: str = "local",
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Publish image from instance.

        :param instance_name: Name of instance to publish image from.
        :param alias: New alias to define at target.
        :param force: Force publishing of image, even if container is running.
        :param image_remote: Name of remote to publish image to.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote instance is found on.

        :raises LXDError: on unexpected error.
        """
        command = [
            "publish",
            f"{remote}:{instance_name}",
            f"{image_remote}:",
        ]

        if alias is not None:
            command.append(f"--alias={alias}")

        if force:
            command.append("--force")

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to publish image from {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def remote_add(
        self, *, remote: str, addr: str, protocol: str = "simplestreams"
    ) -> None:
        """Add a public remote.

        :param remote: Name of remote to add.
        :param addr: Address of remote.
        :param protocol: Name of protocol ("simplestreams" or "lxd").

        :raises LXDError: on unexpected error.
        """
        command = ["remote", "add", remote, addr, f"--protocol={protocol}"]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to add remote {remote!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def remote_list(self) -> Dict[str, Any]:
        """Get list of remotes.

        :returns: dictionary with remote name mapping to config.
        """
        command = ["remote", "list", "--format=yaml"]

        try:
            proc = self._run_lxc(
                command,
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief="Failed to list remotes.",
                details=errors.details_from_called_process_error(error),
            ) from error

        try:
            return load_yaml(proc.stdout)
        except yaml.YAMLError as error:
            raise LXDError(
                brief="Failed to parse lxc remote list.",
                details=(
                    f"* Command that failed: {shlex.join(proc.args)!r}\n"
                    f"* Command output: {proc.stdout!r}"
                ),
            ) from error

    def start(
        self, *, instance_name: str, project: str = "default", remote: str = "local"
    ) -> None:
        """Start container.

        :param instance_name: Name of instance to start.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["start", f"{remote}:{instance_name}"]

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to start {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def stop(
        self,
        *,
        instance_name: str,
        force: bool = False,
        timeout: int = -1,
        project: str = "default",
        remote: str = "local",
    ) -> None:
        """Stop container.

        :param instance_name: Name of instance to stop.
        :param force: Force instance to stop.
        :param timeout: Timeout in seconds. -1 is no timeout.
        :param project: Name of LXD project.
        :param remote: Name of LXD remote.

        :raises LXDError: on unexpected error.
        """
        command = ["stop", f"{remote}:{instance_name}"]

        if force:
            command.append("--force")

        if timeout != -1:
            command.append(f"--timeout={timeout}")

        try:
            self._run_lxc(
                command,
                capture_output=True,
                check=True,
                project=project,
            )
        except subprocess.CalledProcessError as error:
            raise LXDError(
                brief=f"Failed to stop {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Multipass Provider Readiness Checks."""

import logging

from . import errors
from .installer import is_installed
from .multipass import Multipass

logger = logging.getLogger(__name__)


def ensure_multipass_is_ready(*, multipass: Multipass = Multipass()) -> None:
    """Ensure Multipass is ready for use.

    :raises MultipassError: on error.
    """
    if not is_installed():
        raise errors.MultipassError(
            brief="Multipass is required, but not installed.",
            resolution="Visit https://multipass.run for instructions on "
            "installing Multipass for your operating system.",
        )

    if not multipass.is_supported_version():
        version = multipass.version()
        min_version = multipass.minimum_required_version
        raise errors.MultipassError(
            brief=(
                f"Multipass {version!r} does not meet the"
                f" minimum required version {min_version!r}."
            ),
            resolution="Visit https://multipass.run for instructions on "
            "installing Multipass for your operating system.",
        )



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Buildd image(s)."""
import enum
import io
import logging
import pathlib
import re
import subprocess
import sys
import time
from textwrap import dedent
from time import sleep
from typing import Dict, List, Optional, Type

import pydantic
from pydantic import ValidationError

from craft_providers import Base, Executor, errors
from craft_providers.actions import snap_installer
from craft_providers.util.os_release import parse_os_release

from .errors import BaseCompatibilityError, BaseConfigurationError
from .instance_config import InstanceConfiguration

logger = logging.getLogger(__name__)


def default_command_environment() -> Dict[str, Optional[str]]:
    """Provide default command environment dictionary.

    The minimum environment for the buildd image to be configured and function
    properly.  This contains the default environment found in Ubuntu's
    /etc/environment, replaced with the "secure_path" defaults used by sudo for
    instantiating PATH.  In practice it really just means the PATH set by sudo.

    Default /etc/environment found in supported Ubuntu versions:
    PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:
         /usr/bin:/sbin:/bin:/usr/games:/usr/local/games:/snap/bin

    Default /etc/sudoers secure_path found in supported Ubuntu versions:
    PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/snap/bin

    :returns: Dictionary of environment key/values.
    """
    return dict(
        PATH="/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/snap/bin"
    )


def _check_deadline(
    deadline: Optional[float],
    *,
    message: str = "Timed out configuring environment.",
) -> None:
    """Check deadline and raise error if passed.

    :param deadline: Optional time.time() deadline.

    :raises BaseConfigurationError: if deadline is passed.
    """
    if deadline is not None and time.time() >= deadline:
        raise BaseConfigurationError(brief=message)


class BuilddBaseAlias(enum.Enum):
    """Mappings for supported buildd images."""

    XENIAL = "16.04"
    BIONIC = "18.04"
    FOCAL = "20.04"
    JAMMY = "22.04"


class Snap(pydantic.BaseModel, extra=pydantic.Extra.forbid):
    """Details of snap to install in the base.

    :param name: name of snap
    :param channel: snap store channel to install from (default is stable)
      If channel is `None`, then the snap is injected from the host instead
      of being installed from the store.
    :param classic: true if snap is a classic snap (default is false)
    """

    name: str
    channel: Optional[str] = "stable"
    classic: bool = False

    # pylint: disable=no-self-argument
    @pydantic.validator("channel")
    def validate_channel(cls, channel):
        """Validate that channel is not an empty string.

        :raises BaseConfigurationError: if channel is empty
        """
        if channel == "":
            raise BaseConfigurationError(
                brief="channel cannot be empty",
                resolution="set channel to a non-empty string or `None`",
            )
        return channel

    # pylint: enable=no-self-argument


class BuilddBase(Base):
    """Support for Ubuntu minimal buildd images.

    :cvar compatibility_tag: Tag/Version for variant of build configuration and
        setup.  Any change to this version would indicate that prior [versioned]
        instances are incompatible and must be cleaned.  As such, any new value
        should be unique to old values (e.g. incrementing).  It is suggested to
        extend this tag, not overwrite it, e.g.: compatibility_tag =
        f"{appname}-{BuildBase.compatibility_tag}.{apprevision}" to ensure base
        compatibility levels are maintained.
    :cvar instance_config_path: Path to persistent environment configuration
        used for compatibility checks (or other data).  Set to
        /etc/craft-instance.conf, but may be overridden for application-specific
        reasons.
    :cvar instance_config_class: Class defining instance configuration.  May be
        overridden with an application-specific subclass of InstanceConfiguration
        to enable application-specific extensions.

    :param alias: Base alias / version.
    :param environment: Environment to set in /etc/environment.
    :param hostname: Hostname to configure.
    :param snaps: Optional list of snaps to install on the base image.
    :param packages: Optional list of system packages to install on the base image.
    """

    compatibility_tag: str = f"buildd-{Base.compatibility_tag}"
    instance_config_path: pathlib.Path = pathlib.Path("/etc/craft-instance.conf")
    instance_config_class: Type[InstanceConfiguration] = InstanceConfiguration

    def __init__(
        self,
        *,
        alias: BuilddBaseAlias,
        compatibility_tag: Optional[str] = None,
        environment: Optional[Dict[str, Optional[str]]] = None,
        hostname: str = "craft-buildd-instance",
        snaps: Optional[List[Snap]] = None,
        packages: Optional[List[str]] = None,
    ):
        self.alias: BuilddBaseAlias = alias

        if environment is None:
            self.environment = default_command_environment()
        else:
            self.environment = environment

        if compatibility_tag:
            self.compatibility_tag = compatibility_tag

        self._set_hostname(hostname)
        self.snaps = snaps
        self.packages = packages

    def _set_hostname(self, hostname: str) -> None:
        """Set hostname.

        hostname naming convention:
        - between 1 and 63 characters long
        - be made up exclusively of letters, numbers, and hyphens from the ASCII table
        - not begin or end with a hyphen

        If needed, the provided hostname will be trimmed to meet naming conventions.

        :param hostname: hostname to set
        :raises BaseConfigurationError: if the hostname contains no
          alphanumeric characters
        """
        # truncate to 63 characters
        truncated_name = hostname[:63]

        # remove anything that is not an alphanumeric character or hyphen
        name_with_valid_chars = re.sub(r"[^\w-]", "", truncated_name)

        # trim hyphens from the beginning and end
        valid_name = name_with_valid_chars.strip("-")
        if not valid_name:
            raise BaseConfigurationError(
                brief=f"failed to create base with hostname {hostname!r}.",
                details="hostname must contain at least one alphanumeric character",
            )

        logger.debug("Using hostname %r", valid_name)
        self.hostname = valid_name

    def _ensure_instance_config_compatible(
        self, *, executor: Executor, deadline: Optional[float]
    ) -> None:
        """Ensure instance configuration is compatible.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        _check_deadline(deadline)

        try:
            config = InstanceConfiguration.load(
                executor=executor,
                config_path=self.instance_config_path,
            )
        except ValidationError as error:
            raise BaseConfigurationError(
                brief="Failed to parse instance configuration file.",
            ) from error
        # if no config exists, assume base is compatible (likely unfinished setup)
        except FileNotFoundError:
            return

        # make the same assumption as above for empty configs
        if config is None:
            return

        if config.compatibility_tag != self.compatibility_tag:
            raise BaseCompatibilityError(
                reason=(
                    "Expected image compatibility tag "
                    f"{self.compatibility_tag!r}, found {config.compatibility_tag!r}"
                )
            )
        logger.debug(
            "Instance is compatible with compatibility tag %r", config.compatibility_tag
        )

    def _ensure_os_compatible(
        self, *, executor: Executor, deadline: Optional[float]
    ) -> None:
        """Ensure OS is compatible with Base.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        try:
            # Replace encoding errors if it somehow occurs with utf-8. This
            # doesn't need to be perfect for checking compatibility.
            _check_deadline(deadline)
            proc = executor.execute_run(
                command=["cat", "/etc/os-release"],
                capture_output=True,
                check=True,
                text=True,
                encoding="utf-8",
                errors="replace",
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to read /etc/os-release.",
                details=errors.details_from_called_process_error(error),
            ) from error

        os_release = parse_os_release(proc.stdout)

        os_name = os_release.get("NAME")
        if os_name != "Ubuntu":
            raise BaseCompatibilityError(
                reason=f"Expected OS 'Ubuntu', found {os_name!r}"
            )

        compat_version_id = self.alias.value
        version_id = os_release.get("VERSION_ID")
        if version_id != compat_version_id:
            raise BaseCompatibilityError(
                reason=(
                    f"Expected OS version {compat_version_id!r},"
                    f" found {version_id!r}"
                )
            )

    def get_command_environment(
        self,
    ) -> Dict[str, Optional[str]]:
        """Get command environment to use when executing commands.

        :returns: Dictionary of environment, allowing None as a value to
                  indicate that a value should be unset.
        """
        return self.environment.copy()

    def setup(
        self,
        *,
        executor: Executor,
        retry_wait: float = 0.25,
        timeout: Optional[float] = None,
    ) -> None:
        """Prepare base instance for use by the application.

        Wait for environment to become ready and configure it.  At completion of
        setup, the executor environment should have networking up and have all
        of the installed dependencies required for subsequent use by the
        application.

        Setup may be called more than once in a given instance to refresh/update
        the environment.

        If timeout is specified, abort operation if time has been exceeded.

        Guarantees provided by this setup:
          - configured /etc/environment
          - configured hostname
          - networking available (IP & DNS resolution)
          - apt cache up-to-date
          - snapd configured and ready
          - system services are started and ready

        :param executor: Executor for target container.
        :param retry_wait: Duration to sleep() between status checks (if
            required).
        :param timeout: Timeout in seconds.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        if timeout is not None:
            deadline: Optional[float] = time.time() + timeout
        else:
            deadline = None

        self._ensure_os_compatible(executor=executor, deadline=deadline)
        self._ensure_instance_config_compatible(executor=executor, deadline=deadline)
        self._disable_automatic_apt(executor=executor, deadline=deadline)
        self._setup_environment(executor=executor, deadline=deadline)
        self._setup_wait_for_system_ready(
            executor=executor, deadline=deadline, retry_wait=retry_wait
        )
        self._setup_instance_config(executor=executor, deadline=deadline)
        self._setup_hostname(executor=executor, deadline=deadline)
        self._setup_resolved(executor=executor, deadline=deadline)
        self._setup_networkd(executor=executor, deadline=deadline)
        self._setup_wait_for_network(
            executor=executor, deadline=deadline, retry_wait=retry_wait
        )
        self._setup_apt(executor=executor, deadline=deadline)
        self._setup_snapd(executor=executor, deadline=deadline)
        self._setup_snapd_proxy(executor=executor, deadline=deadline)
        self._install_snaps(executor=executor, deadline=deadline)

    def warmup(
        self,
        *,
        executor: Executor,
        retry_wait: float = 0.25,
        timeout: Optional[float] = None,
    ) -> None:
        """Prepare a previously created and setup instance for use by the application.

        Ensure the instance is still valid and wait for environment to become ready.

        Guarantees provided by this wait:
          - OS and instance config are compatible
          - networking available (IP & DNS resolution)
          - system services are started and ready

        If timeout is specified, abort operation if time has been exceeded.

        :param executor: Executor for target container.
        :param retry_wait: Duration to sleep() between status checks (if required).
        :param timeout: Timeout in seconds.

        :raises BaseCompatibilityError: if instance is incompatible.
        :raises BaseConfigurationError: on other unexpected error.
        """
        if timeout is not None:
            deadline: Optional[float] = time.time() + timeout
        else:
            deadline = None

        self._ensure_os_compatible(executor=executor, deadline=deadline)
        self._ensure_instance_config_compatible(executor=executor, deadline=deadline)
        self._setup_wait_for_system_ready(
            executor=executor, deadline=deadline, retry_wait=retry_wait
        )
        self._setup_wait_for_network(
            executor=executor, deadline=deadline, retry_wait=retry_wait
        )
        self._setup_snapd_proxy(executor=executor, deadline=deadline)
        self._install_snaps(executor=executor, deadline=deadline)

    def _disable_automatic_apt(
        self, *, executor: Executor, deadline: Optional[float]
    ) -> None:
        """Disable automatic apt actions.

        This should happen as soon as possible in the instance overall setup,
        to reduce the chances of an automatic apt work being triggered during
        the setup itself (because it includes apt work which may clash
        the triggered unattended jobs).

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        _check_deadline(deadline)
        # set the verification frequency in 10000 days and disable the upgrade
        content = dedent(
            """\
            APT::Periodic::Update-Package-Lists "10000";
            APT::Periodic::Unattended-Upgrade "0";
        """
        ).encode()
        executor.push_file_io(
            destination=pathlib.Path("/etc/apt/apt.conf.d/20auto-upgrades"),
            content=io.BytesIO(content),
            file_mode="0644",
        )

    def _install_snaps(self, *, executor: Executor, deadline: Optional[float]) -> None:
        """Install snaps.

        Snaps will either be installed from the store or injected from the host.
        - If channel is `None` on a linux system, the host snap is injected
          into the provider.
        - If channel is `None` on a non-linux system, an error is raised
          because host injection is not supported on non-linux systems.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        :raises BaseConfigurationError: if the snap cannot be installed
        """
        if not self.snaps:
            logger.debug("No snaps to install.")
            return

        for snap in self.snaps:
            _check_deadline(deadline)
            logger.debug(
                "Installing snap %r with channel=%r and classic=%r",
                snap.name,
                snap.channel,
                snap.classic,
            )

            # don't inject snaps on non-linux hosts
            if sys.platform != "linux" and not snap.channel:
                raise BaseConfigurationError(
                    brief=(
                        f"cannot inject snap {snap.name!r} from host on "
                        "a non-linux system"
                    ),
                    resolution=(
                        "install the snap from the store by setting the "
                        "'channel' parameter"
                    ),
                )

            if snap.channel:
                try:
                    snap_installer.install_from_store(
                        executor=executor,
                        snap_name=snap.name,
                        channel=snap.channel,
                        classic=snap.classic,
                    )
                except snap_installer.SnapInstallationError as error:
                    raise BaseConfigurationError(
                        brief=(
                            f"failed to install snap {snap.name!r} from store"
                            f" channel {snap.channel!r} in target environment."
                        )
                    ) from error
            else:
                try:
                    snap_installer.inject_from_host(
                        executor=executor,
                        snap_name=snap.name,
                        classic=snap.classic,
                    )
                except snap_installer.SnapInstallationError as error:
                    raise BaseConfigurationError(
                        brief=(
                            f"failed to inject host's snap {snap.name!r} "
                            "into target environment."
                        )
                    ) from error

    def _setup_apt(self, *, executor: Executor, deadline: Optional[float]) -> None:
        """Configure apt, update cache and install needed packages.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        _check_deadline(deadline)
        executor.push_file_io(
            destination=pathlib.Path("/etc/apt/apt.conf.d/00no-recommends"),
            content=io.BytesIO('APT::Install-Recommends "false";\n'.encode()),
            file_mode="0644",
        )

        _check_deadline(deadline)
        executor.push_file_io(
            destination=pathlib.Path("/etc/apt/apt.conf.d/00update-errors"),
            content=io.BytesIO('APT::Update::Error-Mode "any";\n'.encode()),
            file_mode="0644",
        )

        try:
            _check_deadline(deadline)
            executor.execute_run(
                ["apt-get", "update"],
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to update apt cache.",
                details=errors.details_from_called_process_error(error),
            ) from error

        # install required packages and user-defined packages
        packages_to_install = ["apt-utils", "curl"]
        if self.packages:
            packages_to_install.extend(self.packages)

        try:
            _check_deadline(deadline)
            executor.execute_run(
                ["apt-get", "install", "-y"] + packages_to_install,
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to install packages.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def _setup_environment(
        self,
        *,
        executor: Executor,
        deadline: Optional[float],
    ) -> None:
        """Configure /etc/environment.

        If environment is None, reset /etc/environment to the default.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        content = (
            "\n".join(
                [f"{k}={v}" for k, v in self.environment.items() if v is not None]
            )
            + "\n"
        ).encode()

        _check_deadline(deadline)
        executor.push_file_io(
            destination=pathlib.Path("/etc/environment"),
            content=io.BytesIO(content),
            file_mode="0644",
        )

    def _setup_hostname(self, *, executor: Executor, deadline: Optional[float]) -> None:
        """Configure hostname, installing /etc/hostname.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        _check_deadline(deadline)
        executor.push_file_io(
            destination=pathlib.Path("/etc/hostname"),
            content=io.BytesIO((self.hostname + "\n").encode()),
            file_mode="0644",
        )

        try:
            _check_deadline(deadline)
            executor.execute_run(
                ["hostname", "-F", "/etc/hostname"],
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to set hostname.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def _setup_instance_config(
        self, *, executor: Executor, deadline: Optional[float]
    ) -> None:
        InstanceConfiguration.update(
            executor=executor,
            data={"compatibility_tag": self.compatibility_tag},
            config_path=self.instance_config_path,
        )
        _check_deadline(deadline)

    def _setup_networkd(self, *, executor: Executor, deadline: Optional[float]) -> None:
        """Configure networkd and start it.

        Installs eth0 network configuration using ipv4.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        _check_deadline(deadline)
        executor.push_file_io(
            destination=pathlib.Path("/etc/systemd/network/10-eth0.network"),
            content=io.BytesIO(
                dedent(
                    """\
                [Match]
                Name=eth0

                [Network]
                DHCP=ipv4
                LinkLocalAddressing=ipv6

                [DHCP]
                RouteMetric=100
                UseMTU=true
                """
                ).encode()
            ),
            file_mode="0644",
        )

        try:
            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "enable", "systemd-networkd"],
                capture_output=True,
                check=True,
            )

            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "restart", "systemd-networkd"],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to setup systemd-networkd.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def _setup_resolved(self, *, executor: Executor, deadline: Optional[float]) -> None:
        """Configure system-resolved to manage resolve.conf.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        try:
            _check_deadline(deadline)
            executor.execute_run(
                [
                    "ln",
                    "-sf",
                    "/run/systemd/resolve/resolv.conf",
                    "/etc/resolv.conf",
                ],
                check=True,
                capture_output=True,
            )

            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "enable", "systemd-resolved"],
                check=True,
                capture_output=True,
            )

            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "restart", "systemd-resolved"],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to setup systemd-resolved.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def _setup_snapd(
        self, *, executor: Executor, deadline: Optional[float] = None
    ) -> None:
        """Install snapd and dependencies and wait until ready.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        try:
            _check_deadline(deadline)
            executor.execute_run(
                [
                    "apt-get",
                    "install",
                    "-y",
                    "fuse",
                    "udev",
                ],
                check=True,
                capture_output=True,
            )

            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "enable", "systemd-udevd"],
                capture_output=True,
                check=True,
            )

            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "start", "systemd-udevd"],
                capture_output=True,
                check=True,
            )

            # This file is created by launchpad-buildd to stop snapd from
            # using the snap store's CDN when running in Canonical's
            # production build farm, since internet access restrictions may
            # prevent it from doing so but will allow the non-CDN storage
            # endpoint.  If this is in place, then we need to propagate it
            # to containers we create.
            no_cdn = pathlib.Path("/etc/systemd/system/snapd.service.d/no-cdn.conf")
            if no_cdn.exists():
                _check_deadline(deadline)
                executor.execute_run(
                    ["mkdir", "-p", no_cdn.parent.as_posix()], check=True
                )

                _check_deadline(deadline)
                executor.push_file(source=no_cdn, destination=no_cdn)

            _check_deadline(deadline)
            executor.execute_run(
                ["apt-get", "install", "-y", "snapd"],
                capture_output=True,
                check=True,
            )

            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "start", "snapd.socket"],
                capture_output=True,
                check=True,
            )

            # Restart, not start, the service in case the environment
            # has changed and the service is already running.
            _check_deadline(deadline)
            executor.execute_run(
                ["systemctl", "restart", "snapd.service"],
                capture_output=True,
                check=True,
            )

            _check_deadline(deadline)
            executor.execute_run(
                ["snap", "wait", "system", "seed.loaded"],
                capture_output=True,
                check=True,
            )

        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to setup snapd.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def _setup_snapd_proxy(
        self, *, executor: Executor, deadline: Optional[float] = None
    ) -> None:
        """Configure the snapd proxy.

        :param executor: Executor for target container.
        :param deadline: Optional time.time() deadline.
        """
        try:
            _check_deadline(deadline)
            http_proxy = self.environment.get("http_proxy")
            if http_proxy:
                command = ["snap", "set", "system", f"proxy.http={http_proxy}"]
            else:
                command = ["snap", "unset", "system", "proxy.http"]
            executor.execute_run(command, capture_output=True, check=True)

            _check_deadline(deadline)
            https_proxy = self.environment.get("https_proxy")
            if https_proxy:
                command = ["snap", "set", "system", f"proxy.https={https_proxy}"]
            else:
                command = ["snap", "unset", "system", "proxy.https"]
            executor.execute_run(command, capture_output=True, check=True)

        except subprocess.CalledProcessError as error:
            raise BaseConfigurationError(
                brief="Failed to set the snapd proxy.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def _setup_wait_for_network(
        self,
        *,
        executor: Executor,
        retry_wait: float = 0.25,
        deadline: Optional[float] = None,
    ) -> None:
        """Wait until networking is ready.

        :param executor: Executor for target container.
        :param retry_wait: Duration to sleep() between status checks.
        :param deadline: Optional time.time() deadline.
        """
        logger.debug("Waiting for networking to be ready...")

        _check_deadline(deadline)
        while True:
            proc = executor.execute_run(
                ["getent", "hosts", "snapcraft.io"],
                capture_output=True,
                check=False,
            )
            if proc.returncode == 0:
                return

            _check_deadline(
                deadline, message="Timed out waiting for networking to be ready."
            )
            sleep(retry_wait)

    def _setup_wait_for_system_ready(
        self,
        *,
        executor: Executor,
        retry_wait: float = 0.25,
        deadline: Optional[float] = None,
    ) -> None:
        """Wait until system is ready.

        :param executor: Executor for target container.
        :param retry_wait: Duration to sleep() between status checks.
        :param deadline: Optional time.time() deadline.
        """
        logger.debug("Waiting for environment to be ready...")

        _check_deadline(deadline)
        while True:
            proc = executor.execute_run(
                ["systemctl", "is-system-running"],
                capture_output=True,
                check=False,
                text=True,
            )

            running_state = proc.stdout.strip()
            if running_state in ["running", "degraded"]:
                return

            logger.debug("systemctl is-system-running status: %s", running_state)

            _check_deadline(
                deadline, message="Timed out waiting for environment to be ready."
            )
            sleep(retry_wait)

    def wait_until_ready(
        self,
        *,
        executor: Executor,
        retry_wait: float = 0.25,
        timeout: Optional[float] = None,
    ) -> None:
        """Wait until base instance is ready.

        Ensure minimum-required boot services are running.  This would be used
        when starting an environment's container/VM after already [recently]
        running setup(), e.g. rebooting the instance.  Allows the environment to
        be used without the cost incurred by re-executing the steps
        unnecessarily.

        If timeout is specified, abort operation if time has been exceeded.

        Guarantees provided by this wait:

            - networking available (IP & DNS resolution)

            - system services are started and ready

        :param executor: Executor for target container.
        :param retry_wait: Duration to sleep() between status checks (if
            required).
        :param timeout: Timeout in seconds.

        :raises ProviderError: on timeout or unexpected error.
        """
        if timeout is not None:
            deadline: Optional[float] = time.time() + timeout
        else:
            deadline = None

        self._setup_wait_for_system_ready(
            executor=executor,
            retry_wait=retry_wait,
            deadline=deadline,
        )
        self._setup_wait_for_network(
            executor=executor,
            retry_wait=retry_wait,
            deadline=deadline,
        )



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""LXD Provider."""

import logging
import os
import shutil
import subprocess
import sys

from craft_providers.errors import details_from_called_process_error

from . import errors
from .lxc import LXC
from .lxd import LXD

logger = logging.getLogger(__name__)


def install(sudo: bool = True) -> str:
    """Install LXD.

    Install application, using sudo if specified.

    :returns: LXD version.

    :raises LXDInstallationError: on installation error.
    :raises LXDError: on unexpected error.
    """
    if sys.platform != "linux":
        raise errors.LXDInstallationError(f"unsupported platform {sys.platform!r}")

    if not sudo and os.geteuid() != 0:
        raise errors.LXDInstallationError("sudo required if not running as root")

    if sudo:
        cmd = ["sudo"]
    else:
        cmd = []

    cmd += ["snap", "install", "lxd"]

    try:
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as error:
        raise errors.LXDInstallationError(
            "Failed to init LXD.",
            details=details_from_called_process_error(error),
        ) from error

    lxd = LXD()
    lxd.wait_ready(sudo=sudo)
    lxd.init(auto=True, sudo=sudo)

    if not is_user_permitted():
        raise errors.LXDInstallationError(
            "user must be manually added to 'lxd' group before using LXD"
        )

    return lxd.version()


def is_initialized(*, remote: str, lxc: LXC) -> bool:
    """Verify that LXD has been initialized and configuration looks valid.

    If LXD has been installed but the user has not initialized it (lxd init),
    the default profile won't have devices configured.  Trying to launch an
    instance or create a project using this profile will result in failures.

    :returns: True if initialized, else False.
    """
    devices = lxc.profile_show(profile="default", remote=remote).get("devices")
    return bool(devices)


def is_installed() -> bool:
    """Check if LXD is installed (and found on PATH).

    :returns: True if lxd is installed.
    """
    return shutil.which("lxd") is not None


def is_user_permitted() -> bool:
    """Check if user has permisisons to connect to LXD.

    :returns: True if user has correct permissions.
    """
    return os.access("/var/snap/lxd/common/lxd/unix.socket", os.O_RDWR)


def ensure_lxd_is_ready(
    *, remote: str = "local", lxc: LXC = LXC(), lxd: LXD = LXD()
) -> None:
    """Ensure LXD is ready for use.

    :raises LXDError: on error.
    """
    if not is_installed():
        raise errors.LXDError(
            brief="LXD is required, but not installed.",
            resolution="Visit https://snapcraft.io/lxd for instructions "
            "on how to install the LXD snap for your distribution.",
        )

    if not lxd.is_supported_version():
        version = lxd.version()
        min_version = lxd.minimum_required_version
        raise errors.LXDError(
            brief=(
                f"LXD {version!r} does not meet the"
                f" minimum required version {min_version!r}."
            ),
            resolution="Visit https://snapcraft.io/lxd for instructions "
            "on how to install the LXD snap for your distribution.",
        )

    if not is_user_permitted():
        raise errors.LXDError(
            brief="LXD requires additional permissions.",
            resolution="Ensure that the user is in the 'lxd' group.",
        )

    if not is_initialized(lxc=lxc, remote=remote):
        raise errors.LXDError(
            brief="LXD has not been properly initialized.",
            resolution="Execute 'lxd init --auto' to initialize LXD.",
        )
